# Purpose:
#   1) Takes sets SCI0 RESOURCE.XXX file(s) & extracts all resources (compressed resources not supported)
#   2) Takes sets of uncompressed 'patch' files & builds RESOURCE.XXX files 

# Primary Use:
#    - Use against RESOURCE.XXX files created with SCI Studio or SCI Companion, which do not use
#      any compression.  
#
# Secondary Usages:
#    - Could be used could to pack new (or replacement) resources into new RESOURCE.XXX files.
#    - One could work around the compression problem by extracting all resources within the Studio or
#      Companion environments, rendering all extracted resources uncompressed.

# Preparation:
#    - For dumping, place RESOURCE.XXX files in the same directory as this script
#    - For packing, create subdirs corresponding to the RESOURCE.XXX files you wish to create.  
#      Place your properly named 'patch' files (e.g. SCRIPT.001, PIC.213) in these directories.

# Usage:    perl resDumpPack.pl dump 
#       OR  perl resDumpPack.pl dumpOrdered
#       OR  perl resDumpPack.pl pack

# Notes:  Compression is NOT supported.  This means two things:
#         1)  Dumping of compressed resources will result in compressed 'patch' resources.  These
#             are useless unless they are decompressed with some other tool.
# 
#         2)  Packing of resources must take place on uncompressed resources, and will not result
#             in compressed resources in the resulting RESOURCE.XXX file

# Do whatever you want with this file, except remove my name.
# Jeremiah Nellis, nellisjp@hotmail.com
# Version 1.0 - Jan 28th, 2012

# SCI0 Resource file specificiations:
#
# The following was taken directly from the FreeSCI documentation:
########################################################################################################
# SCI0 resource.map
#
# The SCI0 map file format is pretty simple: It consists of 6-byte entries, terminated by the sequence
# 0xffff ffff ffff. The first 2 bytes, interpreted as little endian 16 bit integer, encode resource type
# (high 5 bits) and number (low 11 bits). The next 4 bytes are a 32 bit LE integer that contains the 
# resource file number in the high 6 bits, and the absolute offset within the file in the low 26 bits. 
# SCI0 performs a linear search to find the resource; however, multiple entries may match the search, 
# since resources may be present more than once (the inverse mapping is not injective).
#
# SCI0 resource.<nr>
#
# SCI0 resource entries start with a four-tuple of little endian 16 bit words, which we will call 
# (id, comp_size, decomp_size, method). id has the usual SCI0 semantics (high 5 are the resource type, 
# low 11 are its number). comp_size and decomp_size are the size of the compressed and the decompressed
# resource, respectively. The compressed size actually starts counting at the record position of 
# decomp_size, so it counts four bytes in addition to the actual content. method, finally, is the 
# compression method used to store the data.
#
# Resource Storage
# Individual resources can be stored in one of two ways: Either in resource files (which, surprisingly,
# are called something like "RESOURCE.000" or "RESOURCE.001"), or in external patch files (not to be 
# confused with "patch" resources). The external files are called something like "PIC.100" or 
# "script.000", and they take precedence over data from resource files. There is also a file called 
# "RESOURCE.MAP", which contains a lookup table for the individual resources,
# and another file, "RESOURCE.CFG", which contains configuration information.
#
# Resource information stored in external patch files is not compressed and therefore easily readable. 
# It is, however, preceded by two bytes: The first byte contains the resource type ORed with 0x80, the 
# purpose of the second byte is unknown (but it appears to be ignored by the original SCI version 0 
# engine).
#
# As stated before, external patch files take precedence over resource resource files. Applying those
# external files as patches is an option since FreeSCI version 0.2.2.
#
# The resource files, however, are more complicated. Each of them contains a sequence of resources 
# preceded by a header; these resources may be compressed. It is, also, quite common to find resources
# shared by several resource files. The reason for this appears to be that that, back when hard disks
# were rare and hard to come by, the games had to be playable from floppy disks. To prevent unnecessary
# disk-jockeying, common stuff was placed in several resource files, each of which was then stored on one
# disk.
#######################################################################################################

use strict;
use warnings;

use File::Path;
use File::Find;

sub ascii_to_hex($);
sub dumpResource($$$);
sub dumpResources($);
sub packResource($$);
sub packResources();
sub appendResourceFile($$);

if (@ARGV < 1) { 
	  print "\n---------------------------------\n";
      print "Usage: ResDumpPack.pl %1\n";
      print "   %1: dump, dumpOrdered OR pack\n";
	  print "---------------------------------\n\n";
	  print "    dump: Dumps resources into subdirs based on resource file number.\n\n";
	  print "    dumpOrdered: Same as above, but prepends relative position in the resource\n"; 
	  print "                 file to exported file names.  Use this option if you want to\n"; 
	  print "                 re-pack in the same order as the dump.  This option is useful\n";
	  print "                 for validation against the original resource file after a\n";
	  print "                 dump & pack operation.\n\n";
	  print "    pack: Packs resources into resource files based on subdir name.\n";
	  
      exit (1);
}

my $op = $ARGV[0];

if(lc($op) eq "dump") {
   dumpResources(0);
   print "\nDumping resources complete.\n";
} elsif (lc($op) eq "dumpordered") {
   dumpResources(1);
   print "\nDumping resources complete.\n";
} elsif (lc($op) eq "pack") {
   packResources();
   print "\nPacking resources complete.\n";
} else {
   print "Incorrect syntax: use 'dump' or 'pack' as a parameter\n";
}
exit;
 
sub dumpResources($) {
    my $retainOrder = shift;
	# Loop through each resource file
	my $filename = "resource";
	for my $file ( <$filename.*> ) {
		my $resourceNumber = substr($file,index($file,'.') + 1);
		if(length($file) != 12 || !($resourceNumber =~ m/[0-9][0-9][0-9]/) ) {
		   print "Skipping " . $file . "\n";
		} else {
		   print "\n";
		   print "Processing $file\n";
		   print "-----------------------\n";
		   dumpResource($file, $resourceNumber, $retainOrder);
		}
	}
}

sub packResources() {
    use Cwd;
	
	my $workingDir = "work";
	
	if (-e $workingDir) {
	  rmtree($workingDir);
	}

	# this is silly, but we REALLY need to make sure
	# that the dir is gone before we try to recreate it
	rmtree($workingDir);


	mkdir($workingDir) or die "Can't create working directory!\n";
	
    my $dir = getcwd;
	find(  
	     sub {  
		       my $subdir = $_; 
			   if(-d $subdir && $subdir =~ m/[0-9][0-9][0-9]/) {
			      print "\n";
			      print "Processing subdir $subdir\n";
				  print "---------------------\n";
   			      packResource($workingDir,$subdir); 
			   } 
		 	 }, $dir
		);
}



sub ascii_to_hex ($) {
	## Convert each ASCII character to a two-digit hex number.
	(my $str = shift) =~ s/(.|\n)/sprintf("%02lx", ord $1)/eg;
	return $str;
} 

sub dumpResource($$$) {
    my($inputResFile, $resourceNumber, $retainOrder) = @_;
	
	my $subdir = $resourceNumber;
	my $resourceCount = 1;
	
	# Create the subdirectory, named the same as the resource number we are dumping
	# (remove if it already exists)
	if (-e $subdir) {
	  rmtree($subdir);
	}

	# this is silly, but we REALLY need to make sure
	# that the dir is gone before we try to recreate it
	rmtree($subdir);

	mkdir($subdir) or die "Can't create $subdir directory!\n";
	
	# Read whole file in
	my $data = do {
		local $/ = undef;
		open my $fh, $inputResFile
			or die "could not open $inputResFile: $!";
		binmode $fh;
		<$fh>;
	};


	print "Dumping resources...\n";

	my $rest = $data;
	my $strippedFile = $data;

	while ($rest) {

	   # get the 'header' for the resource (8 bytes)
	   #   pattern:  (00 - leading char) | res # - hex + hex 18 through 1B (*256 offset) | comp. size (2) | decomp. size (2) | 00 00 
	   my $hexVal = substr($rest, 0, 8);

	   
	   my $resLen = int(ord(substr($hexVal,2,1))) + int(ord(substr($hexVal,3,1))) * 256 + 4;

	   # strip out only this text resource from the string   
	   my $restText = substr($rest,8,$resLen - 8);
	   
	   $rest = substr($rest,$resLen);  # next character after the length of this resource
	   
	   # strip this resource out of the resource.xxx file
	   my $stripText = quotemeta($hexVal . $restText);
	   $strippedFile =~ s/($stripText)//;

	   my $resourceType = unpack("B32", pack("H*",substr($hexVal,1,1) . substr($hexVal,0,1)) ); 

       my $hex1 = unpack("B16", substr($hexVal,0,1));
	   my $hex2 = unpack("B16", substr($hexVal,1,1));
	   
	   $resourceType = substr($hex2, 0,5);
	   my $resourceNumber = $hex1 . " " . substr($hex2,5);
	   
	   $resourceType = unpack("N", pack("B32", substr("0" x 32 . $resourceType, -32)));
	   my $offset = int(unpack("N", pack("B32", substr("0" x 32 . substr($hex2,5), -32)))) * 256;
	   $resourceNumber = int(unpack("N", pack("B32",substr("0" x 32 . $hex1, -32)))) + $offset;
	   my $compType =  substr(unpack("H2", substr($hexVal,6,1)), 1,1);
  
       my $resourceTypeText;
  
	   if($resourceType eq "0") {  $resourceTypeText = "VIEW"; }
       if($resourceType eq "1") {  $resourceTypeText = "PIC";  }  
	   if($resourceType eq "2") {  $resourceTypeText = "SCRIPT"; }
	   if($resourceType eq "3") {  $resourceTypeText = "TEXT"; }   
	   if($resourceType eq "4") {  $resourceTypeText = "SOUND"; }
	   if($resourceType eq "6") {  $resourceTypeText = "VOCAB"; }
	   if($resourceType eq "7") {  $resourceTypeText = "FONT"; }	  
	   if($resourceType eq "8") {  $resourceTypeText = "CURSOR"; } 
	   if($resourceType eq "9") {  $resourceTypeText = "PATCH"; }
	   
	   # Compression types:  0 - none, 1 - LZW, 2 - COMP-3
 	   print " Resource id: " . uc(ascii_to_hex(substr($hexVal,0,2))) . " Res length = " . $resLen
	   . " resType = " . $resourceType . " resNum = " . $resourceNumber 
	   . " compType = " . $compType . "\n";


	   if($retainOrder == 1) {
	      open(OUTFILE,">" . $subdir . "/" . sprintf("%03d", $resourceCount) . "_" . $resourceTypeText . "." . sprintf("%03d",$resourceNumber)) || die;
	   } else {
	      open(OUTFILE,">" . $subdir . "/" . $resourceTypeText . "." . sprintf("%03d",$resourceNumber)) || die;	   
	   }
       binmode OUTFILE;
	   
	   # write out 2 byte patch header (1st byte 0x80 &'ed with resource type, 2nd byte 0x00)
	   print OUTFILE pack('H*', '8' . $resourceType . '00');
	   
	   
	   # print out data section
	   print OUTFILE $restText;
	   close OUTFILE;
	   
	   $resourceCount++;

	}
}


sub packResource($$) {
    my ($workingDir, $dir) = @_;
    my $subdir = $dir . "/";

	print "Packing resources...\n";
	
	# Loop through each raw resource file
	for my $file ( <$subdir*.*> ) {

		my $resourceNumber = substr($file,index($file,'.') + 1);
		if(!($resourceNumber =~ m/[0-9][0-9][0-9]/)) {
		   #print "Skipping " . $file . "\n";
		} else {
		   appendResourceFile($file, $workingDir . "/RESOURCE." . $dir);
		}  
	}	
}   


sub appendResourceFile ($$) {
    my($inputFile, $outputResFile) = @_;

	# Get the resource type & number from the input file name
	# Consists of two bytes
	# Example: 11011011 00010011
	#                   |||||
	#                  res type = 00010 = 2 (script resource type)
	#    
	#          ||||||||      |||
	#                resource #, take the last 3 bits of the last byte & tack them on the front of the other 8 bits:
	#                            011 + 11011011 = 01111011011 = 987 (resource #)
    #          So, this is script 987.	
	
    # convert the resource number to binary, then to hex
    my $resNumBinary = unpack("B*", pack("N", int(substr($inputFile,index($inputFile,'.') + 1)) ));

	$resNumBinary = substr($resNumBinary, length($resNumBinary) - 11);
	#print "Resource Number (binary) " . $resNumBinary . "\n";

	my $resTypeDec;
	my $resTypeText;
	if(uc($inputFile) =~ m/VIEW/)   { $resTypeDec = 0; $resTypeText = 'VIEW'; }
    if(uc($inputFile) =~ m/PIC/)    { $resTypeDec = 1; $resTypeText = 'PIC'; }
    if(uc($inputFile) =~ m/SCRIPT/) { $resTypeDec = 2; $resTypeText = 'SCRIPT'; }
    if(uc($inputFile) =~ m/TEXT/)   { $resTypeDec = 3; $resTypeText = 'TEXT'; }   
    if(uc($inputFile) =~ m/SOUND/)  { $resTypeDec = 4; $resTypeText = 'SOUND'; }
	# remember, there is no resource #5
    if(uc($inputFile) =~ m/VOCAB/)  { $resTypeDec = 6; $resTypeText = 'VOCAB'; }
    if(uc($inputFile) =~ m/FONT/)   { $resTypeDec = 7; $resTypeText = 'FONT'; }	  
    if(uc($inputFile) =~ m/CURSOR/) { $resTypeDec = 8; $resTypeText = 'CURSOR'; } 
    if(uc($inputFile) =~ m/PATCH/)  { $resTypeDec = 9; $resTypeText = 'PATCH'; }
	
	my $resTypeBinary = unpack("B*", pack("N", $resTypeDec));

	$resTypeBinary = substr($resTypeBinary, length($resTypeBinary) - 5);
	#print "Resource Type (binary) " . $resTypeBinary . "\n";
	
	my $bin1 = substr($resNumBinary, length($resNumBinary) - 8);
	my $bin2 = $resTypeBinary . substr($resNumBinary, 0, 3);
	
	#print $bin1 . "\n";
	#print $bin2 . "\n";
	

	
	my $hex1 = sprintf('%X', oct("0b$bin1"));
	my $hex2 = sprintf('%X', oct("0b$bin2"));
	
	
	
	# pad with leading zero if necessary
    if(length($hex1) == 1) { $hex1 = "0" . $hex1; }
	if(length($hex2) == 1) { $hex2 = "0" . $hex2; }
	

	# Read whole file in
	my $data = do {
		local $/ = undef;
		open my $fh, $inputFile
			or die "could not open $inputFile: $!";
		binmode $fh;
		<$fh>;
	};
	
	# Strip off the first 2 chars (header info);lpp/.                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        
	$data = substr($data, 2);

	my $resourceLength = length($data);

    open RESOURCEFILE, ">>" . $outputResFile || "Can't out $outputResFile $!\n"; 
    binmode RESOURCEFILE;

	# write out resource id
	print RESOURCEFILE pack('H*', $hex1);
	print RESOURCEFILE pack('H*', $hex2);	
	
    # compressed size
   my $compressedLength = int($resourceLength) + int(4); # add length of header + '00' footer
   my $compressedSize1 = $compressedLength % 256;
   my $compressedSize2 = int($compressedLength / 256);
   print RESOURCEFILE pack('C', $compressedSize1);
   print RESOURCEFILE pack('C', $compressedSize2);

    # decompressed size
   my $decompressedLength = int($resourceLength);
   my $decompressedSize1 = $decompressedLength % 256;
   my $decompressedSize2 = int($decompressedLength / 256);
   print RESOURCEFILE pack('C', $decompressedSize1);
   print RESOURCEFILE pack('C', $decompressedSize2);

    # compression type - only support none (0) 
   print RESOURCEFILE pack('H*', '0000');	
	
	#print "Packing Resource file: " . $resourceId . "\n";
	#print "Compressed size: " . $compressedSize1 . " " . $compressedSize2 . "\n";
	#print "Decompressed size: " . $decompressedSize1 . " " . $decompressedSize2 . "\n";
    #print "\n";
	
	print " Resource id: " . $hex1 . $hex2;
	print " Res length = " . (int($resourceLength) + int(8));  # add 8 bytes for header
	print " resType = " . $resTypeText;
	print " resNum = " . substr($inputFile, index($inputFile, '.') + 1);
	print " compType = 0";
	print "\n";
	# write out the file
	print RESOURCEFILE $data;

    close RESOURCEFILE;	
}
